/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.core;

import java.util.ArrayList;

import javax.jcr.PathNotFoundException;
import javax.jcr.RepositoryException;

import org.exoplatform.services.jcr.datamodel.InternalQPath;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: JCRPath.java 5150 2006-04-17 06:44:21Z geaz $
 */

public class JCRPath {

  public static final String ROOT_PATH = "/";

  public static final String SYSTEM_PATH = "/jcr:system";

  private PathElement[] names;

  JCRPath() {
    this.names = new PathElement[0];
  }

  JCRPath addEntry(String namespace, String name, String prefix, int index) {
    addEntry(new PathElement(namespace, name, prefix, index));
    return this;
  }

  JCRPath addEntry(PathElement entry) {
    PathElement[] newNames = new PathElement[names.length + 1];
    for (int i = 0; i < names.length; i++)
      newNames[i] = names[i];
    newNames[names.length] = entry;
    names = newNames;
    return this;
  }

  public JCRPath makeParentPath() {
    return makeAncestorPath(1);
  }

  /**
   * Makes ancestor path by relative degree (For ex relativeDegree == 1 means
   * parent path etc)
   * 
   * @param relativeDegree
   * @return
   * @throws PathNotFoundException
   */
  public JCRPath makeAncestorPath(int relativeDegree) {

    JCRPath path = new JCRPath();
    for (int i = 0; i < names.length - relativeDegree; i++)
      path.addEntry(names[i]);
    return path;
  }
  
  public PathElement[] getRelPath(int relativeDegree) {
    ArrayList entries = new ArrayList();
    for (int i = names.length - relativeDegree; i <names.length ; i++)
      entries.add(names[i]);
    PathElement[] relPath = new PathElement[entries.size()];
    for(int i=0; i<relPath.length; i++)
      relPath[i] = (PathElement)entries.get(i);
    return relPath;
  }

  public InternalQPath getInternalPath() {

    InternalQPath qpath = new InternalQPath();
    for (int i = 0; i < names.length; i++)
      qpath.addEntry(names[i].getNamespace(), names[i].getName(), names[i]
          .getIndex());
    return qpath;
  }

  public String getAsString(boolean showIndex) {
    String path = "";
    if (size() == 1)
      return "/";
    for (int i = 1; i < names.length; i++) {
      path += "/" + names[i].getAsString(showIndex);
    }
    return path;
  }

  public int getDepth() {
    return size() - 1;
  }

  public boolean isDescendantOf(JCRPath ancestorLocation, boolean childOnly) {
    int depthDiff = getDepth() - ancestorLocation.getDepth();
    if (depthDiff <= 0 || (childOnly && depthDiff != 1))
      return false;

    JCRPath.PathElement[] anotherNames = ancestorLocation.getEntries();
    for (int i = 0; i < anotherNames.length; i++) {
      boolean result = anotherNames[i].equals(names[i]);
      if (!result)
        return false;
    }
    return true;
  }

  private int size() {
    return names.length;
  }

  public JCRName getName() {
    return (JCRName) names[size() - 1];
  }

  public int getIndex() {
    return names[size() - 1].getIndex();
  }

  public void setIndex(int index) {
    names[size() - 1].setIndex(index);
  }

  public boolean isIndexSetExplicitly() {
    return names[size() - 1].isIndexSetExplicitly();
  }

  public boolean isSameNameSibling(JCRPath anotherPath) {
    JCRName[] anotherNames = anotherPath.getEntries();
    for (int i = 0; i < anotherNames.length - 1; i++) {
      boolean result = anotherNames[i].equals(names[i]);
      if (!result)
        return false;
    }
    return getName().getName().equals(anotherPath.getName().getName())
        && this.getName().getPrefix().equals(anotherPath.getName().getPrefix());
  }

  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj instanceof JCRPath) {
      JCRPath other = (JCRPath) obj;
      return this.getAsString(true).equals(other.getAsString(true));
    }
    return false;
  }
  

  PathElement[] getEntries() {
    return names;
  }

  public class PathElement extends JCRName {

    private int index;
    private boolean indexSetExplicitly;

    public PathElement(String namespace, String name, String prefix, int index) {
      super(namespace, name, prefix);
      if(index == -1) {
        this.index = 1;
        indexSetExplicitly = false;
      } else {
        this.index = index;
        this.indexSetExplicitly = true;
      }
    }

    public int getIndex() {
      return index;
    }

    public void setIndex(int index) {
      this.index = index;
    }

    public boolean equals(Object obj) {
      if (!(obj instanceof PathElement))
        return false;
      PathElement another = (PathElement) obj;
      return (getAsString(true).equals(another.getAsString(true)) && index == another
          .getIndex());
    }

    public String getAsString(boolean showIndex) {
      String indexStr;
      if (showIndex || getIndex() > 1)
        indexStr = "[" + getIndex() + "]";
      else
        indexStr = "";
      return super.getAsString() + indexStr;
    }

    public boolean isIndexSetExplicitly() {
      return indexSetExplicitly;
    }
  }

}