/**
 **************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jcr.Credentials;
import javax.jcr.LoginException;
import javax.jcr.NamespaceRegistry;
import javax.jcr.NoSuchWorkspaceException;
import javax.jcr.RepositoryException;
import javax.jcr.Session;
import javax.jcr.nodetype.NodeTypeManager;

import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.AuthenticationPolicy;
import org.exoplatform.services.jcr.config.RepositoryConfigurationException;
import org.exoplatform.services.jcr.config.RepositoryEntry;
import org.exoplatform.services.jcr.config.WorkspaceEntry;
import org.exoplatform.services.jcr.core.ManageableRepository;
import org.exoplatform.services.jcr.core.nodetype.ExtendedNodeTypeManager;
import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.impl.RepositoryContainer;
import org.exoplatform.services.jcr.impl.WorkspaceContainer;
import org.exoplatform.services.jcr.impl.core.access.AccessManager;
import org.exoplatform.services.log.ExoLogger;
import org.picocontainer.ComponentAdapter;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: RepositoryImpl.java 5834 2006-05-29 12:05:59Z peterit $
 */
public class RepositoryImpl implements ManageableRepository {
  
  protected static Log log = ExoLogger.getLogger("jcr:RepositoryImpl");

  private static HashMap descriptors = new HashMap();
  static {
    descriptors.put(SPEC_VERSION_DESC, "1.0");
    descriptors.put(SPEC_NAME_DESC, "Content Repository Java Technology API");
    descriptors.put(REP_VENDOR_DESC, "eXo Platform SARL");
    descriptors.put(REP_VENDOR_URL_DESC, "http://www.exoplatform.com");
    descriptors.put(REP_NAME_DESC, "eXo Java Content Repository");
    descriptors.put(REP_VERSION_DESC, "1.0-RC-4");
    descriptors.put(LEVEL_1_SUPPORTED, "true");
    descriptors.put(LEVEL_2_SUPPORTED, "true");
    descriptors.put(OPTION_TRANSACTIONS_SUPPORTED, "false");
    descriptors.put(OPTION_VERSIONING_SUPPORTED, "true");
    descriptors.put(OPTION_OBSERVATION_SUPPORTED, "true");
    descriptors.put(OPTION_LOCKING_SUPPORTED, "true");
    descriptors.put(OPTION_QUERY_SQL_SUPPORTED, "true");
    descriptors.put(QUERY_XPATH_POS_INDEX, "true");
    descriptors.put(QUERY_XPATH_DOC_ORDER, "true");
  }

  private RepositoryContainer repositoryContainer;

  private String systemWorkspaceName;

  //protected Log log;

  //private LogService logService;

  private String name;
  
  private String accessControlType;

  private RepositoryEntry config;
  
  private AuthenticationPolicy authenticationPolicy;

  public RepositoryImpl(RepositoryContainer container)
      throws RepositoryException, RepositoryConfigurationException {
    
    config = (RepositoryEntry) container
        .getComponentInstanceOfType(RepositoryEntry.class);
    
    authenticationPolicy = (AuthenticationPolicy) container
        .getComponentInstanceOfType(AuthenticationPolicy.class);
    
    this.name = config.getName();
    this.systemWorkspaceName = config.getSystemWorkspaceName();
    this.repositoryContainer = container;
    this.accessControlType = config.getAccessControl();
  }

  /**
   * @see javax.jcr.Repository#getDescriptorKeys
   */
  public String[] getDescriptorKeys() {
    String[] keys = new String[descriptors.size()];
    Iterator decriptorsList = descriptors.keySet().iterator();
    int i = 0;
    while (decriptorsList.hasNext())
      keys[i++] = (String) decriptorsList.next();
    return keys;
  }

  /**
   * @see javax.jcr.Repository#getDescriptor
   */
  public String getDescriptor(String key) {
    return (String) descriptors.get(key);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login(Credentials credentials) throws LoginException,
      NoSuchWorkspaceException, RepositoryException {
    return login(credentials, null);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login(String workspaceName) throws LoginException,
      NoSuchWorkspaceException, RepositoryException {
    return login(null, workspaceName);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login() throws LoginException, NoSuchWorkspaceException,
      RepositoryException {
    return login(null, null);
  }

  /**
   * @see javax.jcr.Repository#login
   */
  public Session login(Credentials credentials, String workspaceName)
      throws LoginException, NoSuchWorkspaceException, RepositoryException {

    if (workspaceName == null) {
      workspaceName = config.getDefaultWorkspaceName();
      if (workspaceName == null)
        throw new NoSuchWorkspaceException(
            "Both workspace and default-workspace name are null! ");
    }

    if (!isWorkspaceInitialized(workspaceName))
      throw new NoSuchWorkspaceException(
          "Workspace '"
              + workspaceName
              + "' not found. "
              + "Probably is not initialized. If so either Initialize it manually or turn on the RepositoryInitializer");
    
    WorkspaceContainer workspaceContainer = (WorkspaceContainer) repositoryContainer
        .getComponentInstance(workspaceName);
    
    if (credentials != null)
      return new SessionImpl(workspaceName, (CredentialsImpl) authenticationPolicy.authenticate(credentials),
        workspaceContainer);
    else
      return new SessionImpl(workspaceName, (CredentialsImpl) authenticationPolicy.authenticate(),
        workspaceContainer);
  }
  

  /// -------- ManageableRepository impl -----------------------------
    
  public Session getSystemSession(String workspaceName)
      throws RepositoryException {
    
    WorkspaceContainer wsContainer = getWorkspaceContainer(workspaceName);
    WorkspaceEntry wsConfig = (WorkspaceEntry) wsContainer.getComponentInstanceOfType(WorkspaceEntry.class);
    SessionImpl sysSession = new SessionImpl(wsConfig.getName(), AccessManager.systemCredentials(), wsContainer);
    //return (Session) wsContainer.getComponentInstanceOfType(Session.class);
    return sysSession;
  }

  public RepositoryEntry getConfiguration() {
    return config;
  }
  
  public String[] getWorkspaceNames() {
    List adapters = repositoryContainer
        .getComponentAdaptersOfType(WorkspaceContainer.class);
    String[] workspaceNames = new String[adapters.size()];
    for (int i = 0; i < adapters.size(); i++) {
      ComponentAdapter adapter = (ComponentAdapter) adapters.get(i);
      workspaceNames[i] = new String((String) adapter.getComponentKey());
    }
    return workspaceNames;
  }

  public ExtendedNodeTypeManager getNodeTypeManager() {
    
    return (ExtendedNodeTypeManager) repositoryContainer
        .getComponentInstanceOfType(NodeTypeManager.class);
  }
  
  public NamespaceRegistry getNamespaceRegistry() {
    
    return (NamespaceRegistry) repositoryContainer
        .getComponentInstanceOfType(NamespaceRegistry.class);
  }
  
  public void initWorkspace(String workspaceName, String rootNodeType)
      throws RepositoryException {

    if (isWorkspaceInitialized(workspaceName)) {
      log.warn("Workspace '" + workspaceName + "' is presumably initialized. Initialization canceled");
      return;
    }
    
    InternalQName rootNodeTypeName = getLocationFactory().parseJCRName(rootNodeType).getInternalName();
    NodeData rootNodeData = getWorkspaceInitializer(workspaceName).initWorkspace(rootNodeTypeName);
    
    // TODO register new created workspace in container
        
//    if (isWorkspaceInitialized(workspaceName)) {
//      log.warn("Workspace '" + workspaceName
//          + "' is presumably initialized. Initialization canceled");
//      return;
//    }
//    // System workspace should already been initialized as ordered
//    
//    //if (!isWorkspaceInitialized(systemWorkspaceName)
//    //    && !workspaceName.equals(systemWorkspaceName))
//    //  log.warn("System workspace is not initialized for repository " + name);
//
//    // Init root (/) node
//    Node root = initRootNode(workspaceName, rootNodeType);
//
//    log.info("Root node for " + workspaceName + " initialized. NodeType: "
//        + rootNodeType + " system workspace: " + systemWorkspaceName);
//
//    // Init system workspace
//    if (workspaceName.equals(systemWorkspaceName)) {
//      // initialize /jcr:system
//      Node sys = initJcrSystemNode(root);
//    }

    log.info("Workspace " + workspaceName + " @ " + this.name
        + " is initialized");
  }

  public boolean isWorkspaceInitialized(String workspaceName) {
    try {
      return getWorkspaceInitializer(workspaceName).isWorkspaceInitialized();
    } catch (Exception e) {
      return false;
    }
  }

  ////////////////////////////////////////////////////////

  public RepositoryContainer getContainer() {
    return repositoryContainer;
  }

  public String getSystemWorkspaceName() {
    return systemWorkspaceName;
  }

  public String getName() {
    return name;
  }
  
  
  public SessionImpl getSystemSession() throws RepositoryException {
    return (SessionImpl) getSystemSession(systemWorkspaceName);
  }

  private WorkspaceContainer getWorkspaceContainer(String workspaceName)
      throws RepositoryException {

    WorkspaceContainer wsContainer = (WorkspaceContainer) repositoryContainer
      .getComponentInstance(workspaceName);
    if (wsContainer == null)
      throw new RepositoryException("No container found for workspace "
          + workspaceName + " @ " + this.name);
    return wsContainer;
  }
  
  private WorkspaceInitializer getWorkspaceInitializer(String workspaceName) throws RepositoryException {

    WorkspaceContainer wsContainer = getWorkspaceContainer(workspaceName);
    WorkspaceInitializer wsInitializer = 
      (WorkspaceInitializer) wsContainer.getComponentInstanceOfType(WorkspaceInitializer.class);
    
    if (wsInitializer == null) {
      throw new RepositoryException("No initializer found for workspace container "
          + workspaceName + " @ " + this.name);
    }
    return wsInitializer;
  }

  private LocationFactory getLocationFactory() {
    return (LocationFactory) repositoryContainer
        .getComponentInstanceOfType(LocationFactory.class);
  }
}
