/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.core;

import java.security.AccessControlException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.Property;
import javax.jcr.PropertyIterator;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.RepositoryException;
import javax.jcr.util.TraversingItemVisitor;

import org.apache.commons.collections.MapIterator;
import org.apache.commons.collections.map.ReferenceMap;
import org.apache.commons.logging.Log;
import org.exoplatform.services.jcr.access.PermissionType;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.version.VersionHistoryImpl;
import org.exoplatform.services.jcr.impl.core.version.VersionImpl;
import org.exoplatform.services.jcr.impl.dataflow.persistent.BatcheableWorkspaceDataManager;
import org.exoplatform.services.jcr.impl.dataflow.session.TransactionableDataManager;
import org.exoplatform.services.log.ExoLogger;
import org.exoplatform.services.log.LogService;

/**
 * Created by The eXo Platform SARL .<br>
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: SessionDataManager.java 5834 2006-05-29 12:05:59Z peterit $
 */
public class SessionDataManager {

  protected static Log log = ExoLogger.getLogger("jcr.SessionDataManager");

  private LogService                     logService;

  private SessionImpl                    session;

  private ChangedItemsStorage            changedStorage;

  private SessionMoveStorage             movedStorage;

  // uuid -> ItemImpl
  private Map                            sessionCache;

  private BatcheableWorkspaceDataManager workspaceDataManager;
  
  private TransactionableDataManager transactionableManager;

  public SessionDataManager(SessionImpl session,
      BatcheableWorkspaceDataManager workspaceDataManager, LogService logService)
      throws RepositoryException {
    this.session = session;
    //this.logService = logService;
    // this.log = logService.getLog("org.exoplatform.services.jcr");
    //this.log = (ExoLog) logService.getLog(this.getClass());
    // hard, soft
    this.sessionCache = new ReferenceMap(ReferenceMap.HARD, ReferenceMap.SOFT);
    this.workspaceDataManager = workspaceDataManager;
    this.changedStorage = new ChangedItemsStorage();
    this.movedStorage = new SessionMoveStorage();
    this.transactionableManager = new TransactionableDataManager(workspaceDataManager, session);
  }

  /**
   * Finds item by absolute path in this tnsient storage then in workspace
   * container.
   * 
   * @param path
   * @return item or null if not found
   * @throws RepositoryException
   */
  public ItemImpl getAccessibleItem(JCRPath path) throws RepositoryException {

    ItemImpl item = getItem(path);

    if (item == null || item.getState() == ItemChangeState.DELETED) {
      return null;
    } else if (!item.hasPermission(PermissionType.READ)) {
      throw new AccessDeniedException("Access denied " + path.getAsString(false) + " for "
          + session.getUserID() + " (get item by path)");
    } else {
      return cacheItem(item, false);
    }
  }

  public ItemImpl getItem(JCRPath path) throws RepositoryException {
    // 1. Try in transient changes
    ItemImpl item = changedStorage.get(path);

    if (item == null) {

      // 2. Try from workspacedatamanager
      //ItemData data = workspaceDataManager.getItemData(path);
      ItemData data = transactionableManager.getItemData(path.getInternalPath());
      
      if (data == null)
        return null;
      // parent is lazy loaded
      if (data instanceof NodeData)
        item = createUnchangedNode((NodeData) data);
      else
        item = new PropertyImpl((PropertyData) data, null, ItemChangeState.UNCHANGED, session);
    }

    return item;

  }

  /**
   * @param uuid
   * @return item by UUID or null
   * @throws RepositoryException
   */
  public ItemImpl getAccessibleItemByUUID(String uuid) throws RepositoryException {

    ItemImpl item = getItemByUUID(uuid);

    if (item == null || item.getState() == ItemChangeState.DELETED) {
      return null;
    } else if (!item.hasPermission(PermissionType.READ)) {
      throw new AccessDeniedException("Access denied, item: " + item.getPath()
          + " (get item by UUID)");
    } else {
      log.debug("SessionDataManager found item by UUID " + uuid + " item path: " + item.getPath());
      // [PN] 18.04.06 TODO
      return cacheItem(item, false);
      //return cacheItem(item, true);
    }
  }

  /** [PN] 
   * @param uuid
   * @return item by UUID or null
   * @throws RepositoryException
   */
  public ItemImpl getActualItemByUUID(String uuid) throws RepositoryException {

    ItemImpl item = getItemByUUID(uuid);

    if (item == null || item.getState() == ItemChangeState.DELETED) {
      return null;
      // [PN] 16.02.06 This code deps in cycle with NodeImpl.hasPermisions(), 
      // actualy this method used for reload referenceable node parent
      // which already has a permisions for read
//    } else if (!item.hasPermission(PermissionType.READ)) {
//      throw new AccessDeniedException("Access denied, item: " + item.getPath()
//          + " (get item by UUID)");
    } else {
      log.debug("SessionDataManager found item by UUID " + uuid + " item path: " + item.getPath());

      return cacheItem(item, true); // !!! reload the item
    }
  }  
  
  public ItemImpl getItemByUUID(String uuid) throws RepositoryException {
    // 1. Try in transient changes
    ItemImpl item = changedStorage.get(uuid);

    if (item == null) {

      // 2. Try from workspacedatamanager
      ItemData data = workspaceDataManager.getItemData(uuid);
      if (data == null)
        return null;
      // parent is lazy loaded
      if (data instanceof NodeData) {
        item = createUnchangedNode((NodeData) data);
      } else {
        item = new PropertyImpl((PropertyData) data, null, ItemChangeState.UNCHANGED, session);
      }
    }

    return item;
  }

  private NodeImpl createUnchangedNode(NodeData data) throws RepositoryException {
    NodeImpl node = new NodeImpl(data, null, ItemChangeState.UNCHANGED, session);
//    if (node.isNodeType("nt:version")) {
    if (node.isNodeType(Constants.NT_VERSION)) {
      VersionImpl v = new VersionImpl(data, null, ItemChangeState.UNCHANGED, session);
      return v;
    } else if (node.isNodeType(Constants.NT_VERSION_HISTORY)) {
    //else if (node.isNodeType("nt:versionHistory")) {
      return new VersionHistoryImpl(data, null, ItemChangeState.UNCHANGED, session);
    } else {
      return node;
    }
  }

  List getReferences(String uuid) throws RepositoryException {
    return workspaceDataManager.getReferencesData(uuid);
  }

  /**
   * Mark the changed item as DELETED
   * 
   * @param item
   * @throws RepositoryException
   */
  synchronized void delete(ItemImpl item) throws RepositoryException {
    //log.debug("NodesModificationManager is marking to DELETE " + item.getPath() + " " + item.getInternalUUID());
    
    // Child changes as well
    Iterator<ItemImpl> childChanges = changedStorage.getChanges(item.getLocation(), false).iterator();
    while (childChanges.hasNext()) {
      ItemImpl child = childChanges.next();
      InternalQPath qpath = child.getActualItemData().getQPath();
      if (movedStorage.isMoveSource(qpath) || 
          movedStorage.isMoveSourceDescendant(qpath)) {
        //System.out.println("child moved (DELETED): " + child.getPath());
        child.setState(ItemChangeState.DELETED);
      } else {
        //System.out.println("removed from DELETED: " + child.getPath());
        changedStorage.remove(child);
      }
    }
    //System.out.println("DELETED: " + item.getPath());
    item.setState(ItemChangeState.DELETED);

    changedStorage.put(item);
  }

  /**
   * Adds if not exist or update item in transient storage
   * 
   * @param item
   * @throws RepositoryException
   */
  public synchronized void update(ItemImpl item) throws RepositoryException {

    if (item.getState() != ItemChangeState.ADDED && item.getState() != ItemChangeState.DELETED
        && item.getState() != ItemChangeState.UPDATED
        && item.getState() != ItemChangeState.UNCHANGED) {

      throw new RepositoryException("The state " + ItemChangeState.getStateName(item.getState())
          + " is not accepted");
    }

    ItemImpl item2 = changedStorage.get(item.getInternalUUID());

    if (item2 != null) {
      if (item.isNode() != item2.isNode())
        throw new RepositoryException("There is transient item in " + item.getPath()
            + " with not the same type as incoming. Try to save current first ");
    }

    changedStorage.put(item);
  }
 

  private class ItemsCommitOrderComparator implements Comparator<ItemImpl> {
    public int compare(ItemImpl i1, ItemImpl i2) {
      int stateOrder = i2.getState() - i1.getState();
      int pathOrder = 0;
      int indexOrder = 0;
      if (stateOrder == 0) {
        // For same state order from parent to childs (for ADDED, UPDATED) 
        // or from childs to parent (for DELETED) 
        if (i1.getState() == ItemChangeState.DELETED) {
          pathOrder = i2.getDepth() - i1.getDepth();
        } else {          
          pathOrder = i1.getDepth() - i2.getDepth();                    
        }
        // order by index and name
        try {
          if (pathOrder == 0 && i1.getParent().getPath().equals(i2.getParent().getPath())) {
            if (i1.getName().equals(i2.getName())) {
              // ascending by index
              indexOrder = i1.getLocation().getIndex() - i2.getLocation().getIndex();  
            } else {
              // ascending by name
              indexOrder = i1.getName().compareTo(i2.getName());
              if (indexOrder > 1250) {
                indexOrder = 1250;
              }
            }
          }
        } catch(Exception e) {}        
        pathOrder = pathOrder * 5000 + indexOrder * 2500;        
      } else {
        stateOrder *= 10000;
      }      
      int order = stateOrder + pathOrder;
      //log.info("=== order: " + i1.getPath() + " " + i2.getPath() + " " + order);
      return order;
    }
  }  
  
  private void showCache() {
    for (Object o: sessionCache.values()) {
      if (o instanceof ItemImpl) {
        ItemImpl i = (ItemImpl) o;
        log.info("=== session item: " + i.getPath() + " " + i.getState() + " " + i);
      }
    }    
  }
  
  /**
   * Stored changes (if any) for item and descendants
   * 
   * @param item
   * @throws RepositoryException
   * @throws AccessDeniedException
   * @throws ReferentialIntegrityException
   */
  synchronized void commit(ItemImpl item) throws RepositoryException, AccessDeniedException,
      ReferentialIntegrityException, InvalidItemStateException {   
   
    //showCache();
    
    // 1. validate all, throw an exception if validation failed

    validate(item);
    
    List<ItemImpl> changesList = changedStorage.getChanges(item.getLocation(), false);
    // We have to sort changes list: DELETED items, UPDATED items, ADDED items
    Collections.sort(changesList, new ItemsCommitOrderComparator());
        
    Iterator<ItemImpl> changes = changesList.iterator();

    // 2. create changes list
    List<ItemImpl> resultChanges = new ArrayList<ItemImpl>();
    
    // [PN] 06.05.06
    Map<NodeImpl, Set<String>> reindexChanges = new LinkedHashMap<NodeImpl, Set<String>>();
    Set<NodeImpl> reorderChanges = new LinkedHashSet<NodeImpl>();

    try {

      workspaceDataManager.beginSessionBatch(session);

      while (changes.hasNext()) {
        ItemImpl childItem = changes.next();
        
        boolean reloadState = childItem.isRefreshToPersistedVersion();
        
        try {
          childItem.setRefreshToPersistedVersion(false);
  
          resultChanges.add(childItem);
  
          //log.info("commit() " + ItemChangeState.getStateName(childItem.getState()) + " " + childItem.getPath());
          
          if (childItem.getState() == ItemChangeState.DELETED) {
  
            NodeImpl nodeParent = null;
            if (childItem.isNode()) {
              
              // removing reindex/reorder marks for the item going to delete
              reindexChanges.remove(childItem);
              reorderChanges.remove(childItem);
              
              nodeParent = (NodeImpl) childItem.getParent();
              if (nodeParent != null) {
                // [PN] 06.05.06 mark for reorder
                reorderChanges.add(nodeParent); 
                
                // [PN] 06.05.06 mark for reidex
                String childName = ((NodeImpl) childItem).getName();
                Set<String> reindexNames = reindexChanges.get(nodeParent);
                if (reindexNames != null) {
                  reindexNames.add(childName);
                } else {
                  reindexNames = new LinkedHashSet<String>();
                  reindexNames.add(childName);
                  reindexChanges.put(nodeParent, reindexNames);
                }
              }
            }
            transactionableManager.saveItem(childItem);
            removeFromCache(childItem);
          } else if (childItem.getState() == ItemChangeState.ADDED) {
            
            transactionableManager.saveItem(childItem);
  
            // reset state of children to not to process them one more
            // in a case if they added separately i.e. has ADDED state
            resetItemState(childItem);
          } else if (childItem.getState() == ItemChangeState.UPDATED) {
            
            //System.out.println("UPDATED: "+childItem.getLocation().getAsString(true));
            transactionableManager.saveItem(childItem);
            
            // [PN] 08.02.06 Reset ordered status (TCK tuning)
            if (childItem.isNode()) {
              resetNodeOrderedStatus((NodeImpl) childItem);
            }
          } 
        } finally {
          childItem.setRefreshToPersistedVersion(reloadState);
        }
      }
      
      // Doing actual reindex and reorder of parent childs nodes      
      for (Map.Entry<NodeImpl, Set<String>> reindexTargets: reindexChanges.entrySet()) {
        NodeImpl reindexParent = reindexTargets.getKey();
        Set<String> reindexNames = reindexTargets.getValue();
        
        // Reload childs from storage (node, session changes commited above)
        // and reindex these childs too. 
        reindexParent.setRefreshToPersistedVersion(true);
        for (String childName: reindexNames) {
          reindexParent.reindexChildNodesByName(childName);
        }
        reindexParent.setRefreshToPersistedVersion(false);
      }
      // Reorder and save in the storage (inside reorderChildNodes())
      for (NodeImpl parent: reorderChanges) {
        parent.reorderChildNodes(true);
      }
    } finally {
      workspaceDataManager.endSessionBatch(this.session);
    }

    // 4. clean
    for (int i = 0; i < resultChanges.size(); i++) {

      ItemImpl childItem = resultChanges.get(i);
      // remove from changes table
      changedStorage.remove(childItem);

      // reset state as unchanged
      childItem.setState(ItemChangeState.UNCHANGED);
    }

   
    // 6. clean moved storage
    removeMovedPathDescendants(item.getLocation().getInternalPath());
        
    //showCache();
    
    // 7. [PN] 31.01.06, marking all descendant nodes in session cache for reloading data from workspace 
    refreshCache(item.getLocation());
  }
  
  private void resetNodeOrderedStatus(NodeImpl node) {
    node.setOrderDirection(0);
    // Iterate all childs for doing same, ???
    //NodeIterator childs = node.getNodes();
  }
  
  public void resetItemState(ItemImpl item) throws RepositoryException {
    item.accept(new ItemStateResetter());
  } 
  
  private synchronized void validate(ItemImpl item) throws RepositoryException,
      AccessDeniedException, ReferentialIntegrityException {

    movedStorage.checkViolation(item.getLocation().getInternalPath());

    Iterator<ItemImpl> changes = changedStorage.getChanges(item.getLocation(), false).iterator();
    while (changes.hasNext()) {
      ItemImpl changedItem = changes.next();
      log.debug("SessionDataManager is VALIDATING " + changedItem.getPath() + " "
          + changedItem.getState() + " " + changedItem.getInternalUUID());

      // Check referencial integrity
      if ((changedItem.getState() == ItemChangeState.DELETED || changedItem.getState() == ItemChangeState.AUTO_DELETED)
          && changedItem.isNode() && ((NodeImpl) changedItem).isNodeType("mix:referenceable")) {

        log.debug("SessionDataManager is VALIDATING deleted refereceable " + changedItem.getPath());

        NodeImpl node = (NodeImpl) changedItem;
        List nodeRefs = getReferences(node.getUUID());
        Iterator refs = nodeRefs.iterator();

        while (refs.hasNext()) {
          PropertyData refProp = (PropertyData) refs.next();
          // String path = ((SessionImpl)item.getSession()).
          JCRPath path = session.getLocationFactory().createJCRPath(refProp.getQPath());

          // if ref property is deleted in this session
          int state = getItemState(path);
          if (state == ItemChangeState.DELETED || state == ItemChangeState.AUTO_DELETED)
            continue;

          try {
            session.checkPermission(path.getAsString(false), PermissionType.READ);
          } catch (AccessControlException e) {
            throw new AccessDeniedException("Can not delete node " + node.getPath() + " ("
                + node.getUUID() + ")" + ". It is currently the target of a REFERENCE property "
                + path.getAsString(false) + " ("
                + new String(((ValueData) refProp.getValues().get(0)).getAsByteArray()) + ")");
          }
          throw new ReferentialIntegrityException("Can not delete node " + node.getPath() + " ("
              + node.getUUID() + ")" + ". It is currently the target of a REFERENCE property "
              + path.getAsString(false) + " ("
              + new String(((ValueData) refProp.getValues().get(0)).getAsByteArray()) + ")");
        }
      }

      // validate in workspace data manager
      //workspaceDataManager.validate(changedItem);
    }

  }

  /**
   * Removes all pending changes of this item
   * 
   * @param item
   * @throws RepositoryException
   */
  synchronized void rollback(ItemImpl item) throws InvalidItemStateException, RepositoryException {

    movedStorage.clear();

    Iterator<ItemImpl> changes = changedStorage.getChanges(item.getLocation(), false).iterator();

    while (changes.hasNext()) {
      ItemImpl change = changes.next();

      if (change.isNew()) { // remove from parent
        NodeImpl parent = (NodeImpl) change.getParent();
        if (change.isNode()) {
          if (parent.removeChildNode(change) != null) {
            parent.reorderChildNodes(false);
          }
        } else {
          parent.removeChildProperty(change);
        }
        // should be unchanged?
      } else if (change.isModified()) {
        // [PN]
        // change.loadData(workspaceDataManager.getItemData(change.getLocation()));
      }

      //ItemData itemData = workspaceDataManager.getItemData(change.getLocation()); // [PN]
      ItemData itemData = transactionableManager.getItemData(change.getLocation().getInternalPath());
      if (itemData != null) {
        change.loadData(itemData);
      } else if (item.isNode()) {
        // In loadData ordered status rolled back, but if don't have data reset it manualy
        resetNodeOrderedStatus((NodeImpl) item);
      }
      changedStorage.remove(change);
    }

    // [PN], 29.01.06
    reloadCache(item.getLocation(), false);
  }

  /**
   * Mark all items in cache descendant of jcrPath {@link jcrPath}
   * for refresh with persisted state.  
   * */
  public int refreshCache(JCRPath jcrPath) throws RepositoryException {
    return reloadCache(jcrPath, true);
  }
  
  /**
   * Marks all cached childs nodes for this path {@link jcrPath} for need to reload with 
   * theirs persistent/workspace_cached state, e.g. to be rolled back.
   * 
   * {@link refresh} refresh - 
   * true - then nodes don't removed from cache and root of path don't marked for reload
   * false - then all nodes from this path will be marked, and all items 
   * which has item state no 'unchanged' will be removed from session cache(?)
   * */
  public int reloadCache(JCRPath jcrPath, boolean refresh) throws RepositoryException {

    int reloadedCount = 0;
    List<ItemImpl> values = new ArrayList<ItemImpl>(sessionCache.values());
    for (ItemImpl item : values) {
      // [PN] 26.05.06 Commented for ECM multi-session usage
      //if (item.getLocation().getInternalPath().equals(Constants.JCR_SYSTEM_PATH))
      //  continue;
      //if (item.getLocation().getInternalPath().isDescendantOf(Constants.JCR_SYSTEM_PATH, false))
      //  continue;
      
      //log.info("========== reloadCache() node: " + item.getPath() + "; " + item.getState());
      // If refreshing we don't touch the path root node
      if ((item.getLocation().equals(jcrPath) && !refresh) || item.getLocation().isDescendantOf(jcrPath, false)) {

        // [PN] 31.01.06
        if (!refresh && item.getState() != ItemChangeState.UNCHANGED) {
          //log.info("========== reloadCache() remove from cache: " + item.getPath() + "; " + item.getState());
          item.setRefreshToPersistedVersion(false);
          removeFromCache(item);
          continue;
        }

        // [PN] So, we have to notify all cached childs that parent state is changed,
        // and that they have to load new data
        //log.info("========== reloadCache() mark for reload: " + item.getPath() + "  " + item.getInternalUUID() + " " + item);
        item.setRefreshToPersistedVersion(true);
      }
    }
    return reloadedCount;
  }

  /**
   * Reloads data from persistence storage
   * 
   * @param item
   * @throws RepositoryException
   */
  synchronized void refresh(ItemImpl item) throws InvalidItemStateException, RepositoryException {

    // looking at actual item data for the item in persistent storage
    try {
      if (!isValidItem(item)) {
        throw new InvalidItemStateException("Invalid item object (removed or wrong saved)");
      }
    } catch (InvalidItemStateException e) {
      String message = "This Item object represents a workspace item that has been removed (either by this session).";
      throw new InvalidItemStateException(message + " Cause:[" + e.getMessage() + "] Item: "
          + item.getPath());
    }

    // looping through changes list for items unchanged in this session
    Iterator<ItemImpl> changes = changedStorage.getChanges(item.getLocation(), false).iterator();

    while (changes.hasNext()) {
      ItemImpl change = changes.next();
      if (change.getState() == ItemChangeState.UNCHANGED) {
        //change.loadData(workspaceDataManager.getItemData(change.getLocation()));
        change.loadData(transactionableManager.getItemData(change.getLocation().getInternalPath()));
      }
    }
    
    refreshCache(item.getLocation());
  }

  private boolean isValidItem(ItemImpl item) throws InvalidItemStateException, RepositoryException {

    // check persistent node data
    //ItemData itemData = workspaceDataManager.getItemData(item.getLocation());
    ItemData itemData = transactionableManager.getItemData(item.getLocation().getInternalPath());
    if (itemData == null) {
      String message = "Item has no data in storage for item: ";
      throw new InvalidItemStateException(message + item.getPath());
    }
    // check for delete state marker
    if (item.getState() == ItemChangeState.DELETED
        || item.getState() == ItemChangeState.AUTO_DELETED) {
      String message = "Item has been removed or marked for removing. Item: ";
      throw new InvalidItemStateException(message + item.getPath());
    }
    // check changes for this node
    ItemImpl item2 = changedStorage.get(item.getInternalUUID());
    if (item2 != null
        && item.isNode() == item2.isNode()
        && (item2.getState() == ItemChangeState.DELETED || item2.getState() == ItemChangeState.AUTO_DELETED)) {
      String message = "This Item object represents a workspace item that has been removed (either by this session). Item: ";
      throw new InvalidItemStateException(message + item.getPath());
    }
    boolean result = false;
    if (!item.isRoot()) {
      // So, item is exists and has data, may be smth with parents
      NodeImpl itemParent = (NodeImpl) item.getParent();
      if (itemParent != null) {
        // has parent
        result = isValidItem(itemParent); // recursive
      } // otherwise will throw RepositoryException
    } else {
      result = true; // roos is valid we think that
    }
    return result; // always true, otherwise exception will throw
  }

  /**
   * Returns if node placed in loc or its descendants have pending changes
   * 
   * @param loc
   * @return
   */
  public boolean hasPendingChanges(JCRPath loc, boolean childrenOnly) {
    try {
      Collection items = changedStorage.getChanges(loc, childrenOnly);
      return (items.size() > 0);
    } catch (RepositoryException e) {
      return false;
    }
  }

  public boolean hasAddedOrRemovedDescendant(JCRPath loc, boolean childrenOnly) {
    try {
      Collection<ItemImpl> items = changedStorage.getChanges(loc, childrenOnly);
      for (Iterator<ItemImpl> it = items.iterator(); it.hasNext();) {
        ItemImpl item = it.next();
        if (item.getState() == ItemChangeState.ADDED || item.getState() == ItemChangeState.DELETED)
          return true;
      }
    } catch (RepositoryException e) {
    }
    return false;
  }
  
  /**
   * @author [PN] 09.04.06
   * 
   * For internal use.
   * Returns list of items currently changed in the session and not saved.
   * Will be selected by location and state. 
   * 
   * @param loc 
   * @param itemsState
   * @param childrenOnly
   * @return
   */
  public List<ItemImpl> getChangeStateDescendants(JCRPath loc, int itemsState, boolean childrenOnly) {
    List<ItemImpl> updated = new ArrayList<ItemImpl>();
    try {
      List<ItemImpl> items = changedStorage.getChanges(loc, childrenOnly);
      for (ItemImpl item: items) {
        if (item.getState() == itemsState)
          updated.add(item);
      }
    } catch (RepositoryException e) {
    }
    return updated;
  }

  /**
   * @param node
   * @param isNodes nodes or properties
   * @return changed children of the node
   * @throws RepositoryException
   */
  synchronized List getChildren(NodeImpl node, boolean isNode) throws RepositoryException,
      AccessDeniedException {

    log.debug("SessionDataManager Get children of node " + node.getPath() + " child nodes? "
        + isNode + " " + node.storedNodeData());
    // get items from container
    List<ItemImpl> items = new ArrayList<ItemImpl>();
    
    // [GA] 05.04.06 If the node not new looking for childs in storage 
    if (!node.isNew()) {
      //NodeData data = node.storedNodeData();
      NodeData data = (NodeData) node.getActualItemData();
      if (isNode) {
        // [PN] 24.03.06
      	//List nodes = data.getChildNodes();
        List nodes = workspaceDataManager.getChildNodesData(data);
        if (nodes == null) {
          nodes = new ArrayList();
          log.debug("SessionDataManager Get children of node " + node.getPath() + " nodes == null");
        }
        for (int i = 0; i < nodes.size(); i++) {
          NodeData nodeData = (NodeData) nodes.get(i);
          NodeImpl n = (NodeImpl) getAccessibleItemByUUID(nodeData.getUUID());
          if (n != null) {
            log.debug("SessionDataManager Get child node " + n.getPath());
            items.add(n);
          } else {
            log.debug("CHILD NODE is null " + nodeData.getQPath().getAsString() + " "
                + nodeData.getUUID());
          }
        }
      } else {
        //List props = data.getChildProperties();
        // [PN] 24.03.06
        List props = workspaceDataManager.getChildPropertiesData(data);
        if (props != null) {
          for (int i = 0; i < props.size(); i++) {
            PropertyData propData = (PropertyData) props.get(i);
            PropertyImpl prop = (PropertyImpl) getAccessibleItemByUUID(propData.getUUID());
            if (prop != null)
              items.add(prop);
            else
              log.debug("CHILD PROPERTY is null " + propData.getQPath().getAsString());
          }
        }
      }
    }

    // add ADDED from transient storage ????

    Iterator entries = changedStorage.getChanges().iterator();
    while (entries.hasNext()) {
      ItemImpl item = (ItemImpl) entries.next();
      if (item != null) {
        JCRPath testLocation = item.getLocation();
        if (testLocation.isDescendantOf(node.getLocation(), true)) {
          if (item.isNode() == isNode && item.isNew() && !items.contains(item))
            items.add(item);
        }
      }
    }

    // TODO
    for (int i = 0; i < items.size(); i++) {
      if (items.get(i).getPath().equals(node.getPath()))
        throw new RepositoryException("FATAL: internal error child node path equals parent "
            + node.getPath() + " " + node.getState());

    }

    return items;
  }

  /**
   * Returns state of item in path in session. It includes finding item itself
   * and its ancestor in transient storage In a case if closest ancestor has ADD
   * or DELETE state - this item has it as well.
   * 
   * @param path
   * @return
   */
  int getItemState(JCRPath path) throws RepositoryException {
    // JCRPath path = session.getLocationFactory().createJCRPath(qpath);
    // System.out.println(">>>> getItemState >>>> "+path.getAsString(false)+"
    // "+path.getDepth());

    for (int i = path.getDepth(); i >= 0; i--) {
      ItemImpl item = changedStorage.get(path.makeAncestorPath(i));
      if (item != null) {
        int state = item.getState();
        if (i == 0 || state == ItemChangeState.ADDED || state == ItemChangeState.AUTO_ADDED
            || state == ItemChangeState.DELETED || state == ItemChangeState.AUTO_DELETED) {
          // System.out.println(">>>> getItemState >>>>
          // "+path.getAsString(false)+" "+item.getPath()+" "+state);
          return state;
        }
      }
    }
    // System.out.println(">>>> getItemState >>>> "+path.getAsString(false)+"
    // unchanged");

    return ItemChangeState.UNCHANGED;
  }

  /**
   * insert item into session cache (if not found)
   * 
   * @param item
   * @param reloadState if true item will be reloaded with new state
   * @return inserted item
   * @throws RepositoryException
   */
  public ItemImpl cacheItem(ItemImpl item, boolean reloadState) throws RepositoryException {

//    if (item.getName().equals("jcr:baseVersion")) {
//      log.info("jcr:baseVersion: " + item.getPath());
//    }
    
    String uuid = new String(item.getInternalUUID());

    if (sessionCache.keySet().contains(uuid)) {
      // get value stored in session cache

      ItemImpl storedItem = (ItemImpl) sessionCache.get(uuid);

      if (!reloadState) {
        return storedItem;
      } else {
        storedItem.loadData(item.getActualItemData());
        storedItem.setState(item.getState());
        item = storedItem;
      }
    }

    // put item to session cache
    sessionCache.put(uuid, item);
    // System.out.println("SESSION CACHE put "+uuid+" "+item.getPath()+"
    // "+item);

    if (item.isNode()) {
      NodeImpl node = (NodeImpl) item;
      if (node.isChildPropertiesInitialized()) { // .arePropertiesInitialized()
        PropertyIterator props = node.getProperties();
        while (props.hasNext())
          cacheItem((PropertyImpl) props.nextProperty(), reloadState);
      }

      if (node.isChildNodesInitialized()) { // .areNodesInitialized()
        NodeIterator nodes = node.getNodes();
        while (nodes.hasNext())
          cacheItem((NodeImpl) nodes.nextNode(), reloadState);
      }
    }

    return item;
  }

  public void removeFromCache(ItemImpl item) {
        
    if (item == null) return;
    
    if(item.getLocation().getInternalPath().isDescendantOf(Constants.JCR_SYSTEM_PATH, false)) 
      return;
    
    // Remove from ReferenceMap, doing it instead GC
    if (sessionCache instanceof ReferenceMap) {
      ReferenceMap cache = (ReferenceMap) sessionCache;

      MapIterator cacheIterator = cache.mapIterator();
      List<String> childsForRemove = new ArrayList<String>();
      while (cacheIterator.hasNext()) {
        String cachedUuid = (String) cacheIterator.next();
        //ItemImpl cachedItem = (ItemImpl) cache.get(cUuid); 
        ItemImpl cachedItem = (ItemImpl) cacheIterator.getValue();
        if (cachedItem.isNode()) {
          if (cachedItem.getLocation().isDescendantOf(item.getLocation(), false)) {
            //System.err.println("===== removeFromCache() " + cachedItem.getPath() + "  " + cachedUuid);
            childsForRemove.add(cachedUuid);
            continue;
          }
        }
        if (item.getInternalUUID().equals(cachedUuid)) {                    
          cacheIterator.remove();
        }
      }
      // Removing childs of the item
      cacheIterator = cache.mapIterator();
      for (String uuid: childsForRemove) {
        while (cacheIterator.hasNext()) {
          String cachedUuid = (String) cacheIterator.next();
          if (uuid.equals(cachedUuid)) {
            cacheIterator.remove();
          }
        }
      }
    }
  }

/**
   * @return Returns the workspDataManager.
   */
  public BatcheableWorkspaceDataManager getWorkspaceDataManager() {
    return workspaceDataManager;
  }

  public String dump() {
    String d = "\nChanges:";
    for (Iterator changes = changedStorage.getChanges().iterator(); changes.hasNext();) {
      ItemImpl item = (ItemImpl) changes.next();
      d += "\nItemImpl: isNode:" + item.isNode() + " uuid: " + item.getInternalUUID() + " path:"
          + item.getPath() + " state:" + ItemChangeState.getStateName(item.getState());
    }

    d += "\nCache:";
    for (Iterator cache = sessionCache.values().iterator(); cache.hasNext();) {
      ItemImpl item = (ItemImpl) cache.next();
      d += "\nItemImpl: isNode:" + item.isNode() + " uuid: " + item.getInternalUUID() + " path:"
          + item.getPath() + " state:" + ItemChangeState.getStateName(item.getState());
    }

    return d;
  }

  private class ItemStateResetter extends TraversingItemVisitor.Default {

    protected void entering(Node node, int level) throws RepositoryException {
      NodeImpl nodeImpl = (NodeImpl) node;
      nodeImpl.setState(ItemChangeState.UNCHANGED);
      // System.out.println("RESET>>>
      // "+sessionCache.get(nodeImpl.getInternalUUID())+" "+nodeImpl);
    }

    protected void entering(Property property, int level) throws RepositoryException {
      ((ItemImpl) property).setState(ItemChangeState.UNCHANGED);
    }

  }

  void addMovedPaths(InternalQPath path1, InternalQPath path2) throws RepositoryException {
    movedStorage.addMovedPaths(path1, path2);
  }
  
  void removeMovedPathDescendants(InternalQPath path) {
    movedStorage.removeDescendantsIfAny(path);
  }

  /**
   * @return Returns the movedStorage.
   */
  public SessionMoveStorage getMovedStorage() {
    return movedStorage;
  }
  
  public TransactionableDataManager getTransactManager() {
    return transactionableManager;
  }
}
