/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.apache.commons.logging.Log;
import org.exoplatform.commons.utils.ISO8601;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;
import org.exoplatform.services.log.ExoLogger;

/**
 * A <code>DateValue</code> provides an implementation of the
 * <code>Value</code> interface representing a date value.
 * 
 * @author Stefan Guggisberg
 */
public class DateValue extends BaseValue {

  public static final int TYPE = PropertyType.DATE;
  
  public static final DateFormat DATETZ_FORMAT = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ");
  
  protected static final String CALENDAR_FIELDS_DELIMITER = ";"; // [PN] I hope it's unique will be for any time zone ID etc.
  protected static final String CALENDAR_FIELDS_SEPARATOR = "--";

  protected static Log log = ExoLogger.getLogger("jcr.DateValue");
  
  private final Calendar  date;

  /**
   * Constructs a <code>DateValue</code> object representing a date.
   * 
   * @param date the date this <code>DateValue</code> should represent s
   */
  public DateValue(Calendar date) {
    super(TYPE, new TransientValueData(date));
    this.date = (Calendar) date.clone();
  }

  DateValue(ValueData data) {
    super(TYPE, data);
    //this.date = ISO8601.parse(new String(data.getAsByteArray()));
    this.date = deserializeDate(data.getAsByteArray());
  }
  
  /**
   * Serializing calendar date in string containing 
   * ISO8601 formated data (YYYY-MM-DDThh:mm:ss.SSSTZD)
   * plus information about specific java.util.Calendar fields 
   * (isLenient, getFirstDayOfWeek, getMinimalDaysInFirstWeek and getTimeZone).
   * 
   * @param date
   * @return
   */
  static public byte[] serializeDate(Calendar date) {
    
    // String isoString = ISO8601.format(date);
    String calendarString = CALENDAR_FIELDS_SEPARATOR 
      + date.isLenient() + CALENDAR_FIELDS_DELIMITER 
      + date.getFirstDayOfWeek() + CALENDAR_FIELDS_DELIMITER 
      + date.getMinimalDaysInFirstWeek() + CALENDAR_FIELDS_DELIMITER 
      + date.getTimeZone().getID();
    // String sd = isoString + calendarString;
    String sd = DATETZ_FORMAT.format(date.getTime()) + calendarString;
    return sd.getBytes(); 
  }
  
  
  /**
   * @deprecated
   */
  static public byte[] serializeCalendar(Calendar date) {
    if (date == null) {
      throw new IllegalArgumentException("argument can not be null");
    }

    ByteArrayOutputStream baos = new ByteArrayOutputStream();
    ObjectOutputStream out = null;
    try {
      out = new ObjectOutputStream(baos);
      out.writeObject(date);
      out.close();
      return baos.toByteArray();
    } catch (Exception e) {
      log.error("Error serialize date value (Calendar) from object " + date
          + ". Error: " + e);
    }
    return null;
  }

  /**
   * Deserializing calendar date from string containing 
   * ISO8601 formated data (YYYY-MM-DDThh:mm:ss.SSSTZD)
   * plus information about specific java.util.Calendar fields 
   * (isLenient, getFirstDayOfWeek, getMinimalDaysInFirstWeek and getTimeZone).
   * 
   * @param dateBytes
   * @return
   */
  static public Calendar deserializeDate(byte[] dateBytes) {
    
    String serString = null; 
    try {
      serString = new String(dateBytes, DEFAULT_ENCODING);
    } catch(UnsupportedEncodingException e) {
      log.warn("Error deserialize date value (Calendar) from bytes '"
          + new String(dateBytes) + "' with DEFAULT_ENCODING " + DEFAULT_ENCODING, e);
      serString = new String(dateBytes);
    }
    String[] parts = serString.split(CALENDAR_FIELDS_SEPARATOR);
    if (parts.length == 2) {
      try {
        //Calendar isoDate = ISO8601.parse(parts[0]);
        Date isoDate = DATETZ_FORMAT.parse(parts[0]);
        Calendar isoCalendar = Calendar.getInstance();
        isoCalendar.setTime(isoDate);
        String[] calendarFileds = parts[1].split(CALENDAR_FIELDS_DELIMITER);
        if (calendarFileds.length == 4) {
          // Construct new time zone using offset and ID
          //          try {
          //            TimeZone tzInfo = TimeZone.getTimeZone(calendarFileds[3]);
          //            //tzInfo.setRawOffset(isoDate.getTimeZone().getOffset(isoDate.getTimeInMillis()));
          //            tzInfo.setRawOffset(isoCalendar.getTimeZone().getRawOffset());
          //            isoCalendar.setTimeZone(tzInfo);
          //          } catch(Exception e) {
          //            log.warn("Can't parse and init serialized time zone '" + calendarFileds[3] + "'", e);
          //          }
          try {
            isoCalendar.setLenient(Boolean.parseBoolean(calendarFileds[0]));
            isoCalendar.setFirstDayOfWeek(Integer.parseInt(calendarFileds[1]));
            isoCalendar.setMinimalDaysInFirstWeek(Integer.parseInt(calendarFileds[2]));
          } catch(Exception e) {
            log.warn("Can't parse serialized fields for the calendar '" 
                + parts[1] + "' but calendar has '" + isoDate.toString() + "'", e);
          }
        } else {
          log.warn("Fields of the calendar is not serialized properly '" 
              + parts[1] + "' but calendar has '" + isoDate.toString() + "'");
        }
        return isoCalendar;
      } catch(ParseException e) {
        log.warn("Can't parse serialized calendar '" + parts[0] + "'", e);
      }
    }
    return null;
  }
  
  /**
   * @deprecated
   */
  static public Calendar deserializeCalendar(byte[] dateBytes) {
    if (dateBytes == null) {
      throw new IllegalArgumentException("argument can not be null");
    }
    
    ByteArrayInputStream bais = new ByteArrayInputStream(dateBytes);
    ObjectInputStream in = null;
    try {
      in = new ObjectInputStream(bais);
      Object o = in.readObject();
      in.close();
      return o instanceof Calendar ? (Calendar) o : null;
    } catch (Exception e) {
      log.error("Error deserialize date value (Calendar) from bytes '"
          + new String(dateBytes) + "'. Error: " + e);
    }
    return null;
  }

  /**
   * Indicates whether some other object is "equal to" this one.
   * <p>
   * The result is <code>true</code> if and only if the argument is not
   * <code>null</code> and is a <code>DateValue</code> object that
   * represents the same value as this object.
   * 
   * @param obj the reference object with which to compare.
   * @return <code>true</code> if this object is the same as the obj argument;
   *         <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj instanceof DateValue) {
      DateValue other = (DateValue) obj;
      if (date == other.date) {
        return true;
      } else if (date != null && other.date != null) {
        return date.equals(other.date);
      }
    }
    return false;
  }

  /**
   * @see BaseValue#getInternalString()
   */
  protected String getInternalString() throws ValueFormatException {
    if (date != null) {
      return ISO8601.format(date);
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  // ----------------------------------------------------------------< Value >
  /**
   * @see Value#getDate
   */
  public Calendar getDate() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    if (date != null) {
      return date;
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  /**
   * @see Value#getLong
   */
  public long getLong() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    if (date != null) {
      return date.getTimeInMillis();
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  /**
   * @see Value#getBoolean
   */
  public boolean getBoolean() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    if (date != null) {
      throw new ValueFormatException("cannot convert date to boolean");
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  /**
   * @see Value#getDouble
   */
  public double getDouble() throws ValueFormatException, IllegalStateException, RepositoryException {
    setValueConsumed();

    if (date != null) {
      long ms = date.getTimeInMillis();
      if (ms <= Double.MAX_VALUE) {
        return ms;
      }
      throw new ValueFormatException("conversion from date to double failed: inconvertible types");
    } else {
      throw new ValueFormatException("empty value");
    }
  }
}
