/*
 * Copyright 2005 Day Management AG, Switzerland. All rights reserved.
 */
package org.exoplatform.services.jcr.impl.core.value;

import java.util.Calendar;

import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;

import org.exoplatform.services.jcr.datamodel.IllegalPathException;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.core.JCRPath;
import org.exoplatform.services.jcr.impl.core.LocationFactory;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;

/**
 * A <code>PathValue</code> provides an implementation of the
 * <code>Value</code> interface representing a <code>PATH</code> value (an
 * absolute or relative workspace path).
 * 
 * @author Stefan Guggisberg
 */
public class PathValue extends BaseValue {

  public static final int TYPE = PropertyType.PATH;

  private final JCRPath path;

  /**
   * Protected constructor creating a <code>PathValue</code> object without
   * validating the path.
   * 
   * @param path
   *          the path this <code>PathValue</code> should represent
   * @see #valueOf
   */
  PathValue(JCRPath path) throws RepositoryException {
    super(TYPE, new TransientValueData(path.getInternalPath()));
    this.path = path;
  }

  PathValue(ValueData data, LocationFactory locationFactory)
      throws RepositoryException {
    super(TYPE, data);
    InternalQPath qpath;
    try {
      qpath = InternalQPath.parse(new String(data.getAsByteArray()));
    } catch (IllegalPathException e) {
      throw new RepositoryException(e);
    }
    this.path = locationFactory.createJCRPath(qpath);
  }

  /**
   * Indicates whether some other object is "equal to" this one. <p/>The result
   * is <code>true</code> if and only if the argument is not <code>null</code>
   * and is a <code>PathValue</code> object that represents the same value as
   * this object.
   * 
   * @param obj
   *          the reference object with which to compare.
   * @return <code>true</code> if this object is the same as the obj argument;
   *         <code>false</code> otherwise.
   */
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (obj instanceof PathValue) {
      PathValue other = (PathValue) obj;
      if (path == other.path) {
        return true;
      } else if (path != null && other.path != null) {
        return path.equals(other.path);
      }
    }
    return false;
  }

  /**
   * @see BaseValue#getInternalString
   */
  protected String getInternalString() throws ValueFormatException {
    if (path != null) {
      return path.getAsString(false);
    } else {
      throw new ValueFormatException("empty value");
    }
  }

  //----------------------------------------------------------------< Value >
  /**
   * @see Value#getDate
   */
  public Calendar getDate() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to date failed: inconvertible types");
  }

  /**
   * @see Value#getLong
   */
  public long getLong() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to long failed: inconvertible types");
  }

  /**
   * @see Value#getBoolean
   */
  public boolean getBoolean() throws ValueFormatException,
      IllegalStateException, RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to boolean failed: inconvertible types");
  }

  /**
   * @see Value#getDouble
   */
  public double getDouble() throws ValueFormatException, IllegalStateException,
      RepositoryException {
    setValueConsumed();

    throw new ValueFormatException(
        "conversion to double failed: inconvertible types");
  }
}