/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.core.value;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Calendar;

import javax.jcr.Node;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFactory;
import javax.jcr.ValueFormatException;

import org.exoplatform.commons.utils.ISO8601;
import org.exoplatform.services.jcr.core.ExtendedPropertyType;
import org.exoplatform.services.jcr.datamodel.Uuid;
import org.exoplatform.services.jcr.datamodel.ValueData;
import org.exoplatform.services.jcr.impl.core.JCRName;
import org.exoplatform.services.jcr.impl.core.JCRPath;
import org.exoplatform.services.jcr.impl.core.LocationFactory;
import org.exoplatform.services.jcr.impl.dataflow.TransientValueData;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov </a>
 * @version $Id: ValueFactoryImpl.java 5623 2006-05-19 16:13:09Z peterit $
 */

public class ValueFactoryImpl implements ValueFactory {

  private LocationFactory locationFactory;

  public ValueFactoryImpl(LocationFactory locationFactory) {
    this.locationFactory = locationFactory;
  }

  /*
   * @see javax.jcr.ValueFactory#createValue(java.lang.String, int)
   */
public Value createValue(String value, int type) throws ValueFormatException {
    try {
      switch (type) {
      case PropertyType.STRING:
        return createValue(new String(value));
      case PropertyType.BINARY:
        // Warning: applicable if there ir real string not binary data
        return createValue(new ByteArrayInputStream(value.getBytes()));
      case PropertyType.BOOLEAN:
        return createValue(Boolean.parseBoolean(value));
      case PropertyType.LONG:
        return createValue(Long.parseLong(value));
      case PropertyType.DOUBLE:
        return createValue(Double.parseDouble(value));
      case PropertyType.DATE:
        Calendar cal = ISO8601.parse(value);
        if (cal == null) // illegal value - parse returns null
          throw new ValueFormatException("Date '" + value + "' is invalid");
        return createValue(cal);
      case PropertyType.PATH:
        try {
          JCRPath path;
          if (value.startsWith("/"))
            path = locationFactory.parseAbsPath(value);
          else
            path = locationFactory.parseRelPath(value);
          // return new PathValue(path);
          return createValue(path);
        } catch (RepositoryException e) {
          throw new ValueFormatException("Path '" + value + "' is invalid");
        }
      case PropertyType.NAME:
        try {
          JCRName name = locationFactory.parseJCRName(value);
          // return new NameValue(name);
          return createValue(name);
        } catch (RepositoryException e) {
          throw new ValueFormatException("Name '" + value + "' is invalid", e);
        }
      case PropertyType.REFERENCE:
        // return new ReferenceValue(new Uuid(value));
        return createValue(new Uuid(value));
      case ExtendedPropertyType.PERMISSION:
        return PermissionValue.parseValue(value);
      default:
        throw new ValueFormatException("unknown type " + type);
      }
    } catch (IllegalArgumentException e) { //NumberFormatException
      throw new ValueFormatException("Cant create value from string '"
          + value + "' for type " + PropertyType.nameFromValue(type));
    }
  }
  /*
   * @see javax.jcr.ValueFactory#createValue(java.lang.String)
   */
  public Value createValue(String value) {
    if (value == null)
      return null;
    return new StringValue(value);
  }

  /*
   * @see javax.jcr.ValueFactory#createValue(long)
   */
  public Value createValue(long value) {
    return new LongValue(value);
  }

  /*
   * @see javax.jcr.ValueFactory#createValue(double)
   */
  public Value createValue(double value) {
    return new DoubleValue(value);
  }

  /*
   * @see javax.jcr.ValueFactory#createValue(boolean)
   */
  public Value createValue(boolean value) {
    return new BooleanValue(value);
  }

  /*
   * @see javax.jcr.ValueFactory#createValue(java.util.Calendar)
   */
  public Value createValue(Calendar value) {
    if (value == null)
      return null;
    return new DateValue(value);
  }

  /*
   * @see javax.jcr.ValueFactory#createValue(java.io.InputStream)
   */
  public Value createValue(InputStream value) {
    if (value == null)
      return null;
    return new BinaryValue(value);
  }

  /*
   * @see javax.jcr.ValueFactory#createValue(javax.jcr.Node)
   */
  public Value createValue(Node value) throws RepositoryException {
    if (value == null)
      return null;
    if (!value.isNodeType("mix:referenceable"))
      throw new ValueFormatException("Node " + value.getPath() + " is not referenceable");
    return new ReferenceValue(value);
  }

  // /////////////////////////

  public Value createValue(JCRName value) throws RepositoryException {
    if (value == null)
      return null;
    return new NameValue(value);
  }

  public Value createValue(JCRPath value) throws RepositoryException {
    if (value == null)
      return null;
    return new PathValue(value);
  }

  public Value createValue(byte[] value) {
    return new BinaryValue(value);
  }

  public Value createValue(Uuid value) {
    if (value == null)
      return null;
    return new ReferenceValue(value);
  }

  public Value loadValue(ValueData data, int type) throws RepositoryException {

    // byte[] value = data.getAsByteArray();

    if (data.getAsByteArray() == null) // possible for multivalued ?
      return null;

    switch (type) {
    case PropertyType.STRING:
      return new StringValue(data);
    case PropertyType.BINARY:
      return new BinaryValue(data);
    case PropertyType.BOOLEAN:
      return new BooleanValue(data);
    case PropertyType.LONG:
      return new LongValue(data);
    case PropertyType.DOUBLE:
      return new DoubleValue(data);
    case PropertyType.DATE:
      return new DateValue(data);
    case PropertyType.PATH:
      return new PathValue(data, locationFactory);
    case PropertyType.NAME:
      return new NameValue(data, locationFactory);
    case PropertyType.REFERENCE:
      return new ReferenceValue(data);
    case PropertyType.UNDEFINED:
      return null;
    case ExtendedPropertyType.PERMISSION:
      return new PermissionValue(data);
    default:
      throw new ValueFormatException("unknown type " + type);

    }
  }

  public Value[] convert(Value[] oldValues, int type, boolean forceCopy)
      throws RepositoryException, ValueFormatException {
    Value[] newValues = new Value[oldValues.length];
    for (int i = 0; i < oldValues.length; i++)
      newValues[i] = convert(oldValues[i], type, forceCopy);
    return newValues;
  }

  public Value convert(Value oldValue, int type, boolean forceCopy) throws RepositoryException,
      ValueFormatException {

    if (oldValue == null)
      return oldValue;
    else if (oldValue.getType() == type && !forceCopy)
      return oldValue;

    // do not rely on BaseValue!
    //BaseValue value = (BaseValue) oldValue;
    Value value = oldValue;

    switch (type) {
    case PropertyType.STRING:
      return createValue(value.getString());
    case PropertyType.BINARY:
      //return createValue(value.getData().getAsByteArray());
      return createValue(value.getStream());
    case PropertyType.BOOLEAN:
      return createValue(value.getBoolean());
    case PropertyType.LONG:
      return createValue(value.getLong());
    case PropertyType.DOUBLE:
      return createValue(value.getDouble());
    case PropertyType.DATE:
      if (value instanceof DateValue) {
        return createValue(((DateValue) value).getDate());
      } else {
        return createValue(value.getDate());
      }
    case PropertyType.PATH:
      String stringValue = value.getString();
      JCRPath pathValue = null;
      try {
        // trying relPath before
        pathValue = locationFactory.parseRelPath(stringValue);
      } catch (RepositoryException e) {
        // so, may be it's a abs path
        try {
          pathValue = locationFactory.parseAbsPath(stringValue);
        } catch (RepositoryException e1) {
          throw new RepositoryException("Can't convert to absolute PATH value. "
              + " Convert to relative PATH value is impossible too (" + e.getMessage() + ")", e1);
        }
      }
      return createValue(pathValue);
    case PropertyType.NAME:
      return createValue(locationFactory.parseJCRName(value.getString()));
    case PropertyType.REFERENCE:
      //return createValue(value.getReference(), PropertyType.REFERENCE);
      return createValue(value.getString(), PropertyType.REFERENCE);
    case PropertyType.UNDEFINED:
      return value;
    case ExtendedPropertyType.PERMISSION:
      PermissionValue permValue = (PermissionValue) value;
      return new PermissionValue(permValue.getIdentity(), permValue.getPermission());
    default:
      throw new ValueFormatException("ValueFactory.convert() unknown or unconvertable type " + type);
    }
  }

  private String streamToString(InputStream in) {
    try {
      in.mark(in.available());
      ByteArrayOutputStream out = new ByteArrayOutputStream();
      byte[] buffer = new byte[8192];
      int read;
      while ((read = in.read(buffer)) > 0) {
        out.write(buffer, 0, read);
      }
      in.reset();
      return new String(out.toByteArray(), BaseValue.DEFAULT_ENCODING);
    } catch (IOException e) {
      e.printStackTrace();
      return null;
    }
  }
  
}