/**
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 */

package org.exoplatform.services.jcr.impl.dataflow.persistent;

import java.util.ArrayList;
import java.util.List;

import javax.jcr.InvalidItemStateException;
import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.ValueFactory;

import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.datamodel.PropertyData;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.ItemChangeState;
import org.exoplatform.services.jcr.impl.core.ItemImpl;
import org.exoplatform.services.jcr.impl.core.ItemState;
import org.exoplatform.services.jcr.impl.core.JCRPath;
import org.exoplatform.services.jcr.impl.core.NodeImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.core.version.VersionHistoryImpl;
import org.exoplatform.services.jcr.impl.core.version.VersionImpl;
import org.exoplatform.services.jcr.impl.dataflow.DataManager;
import org.exoplatform.services.jcr.impl.dataflow.ItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.TransientItemData;
import org.exoplatform.services.jcr.impl.dataflow.TransientNodeData;
import org.exoplatform.services.jcr.storage.WorkspaceDataContainer;

/**
 * Created by The eXo Platform SARL        .
 * 
 * Responsible for: 
 * *redirecting repository operations if item is descendant of
 * /jcr:system/jcr:versionStorage
 * *adding version history for newly added/assigned mix:versionable 
 * 
 * @author <a href="mailto:gennady.azarenkov@exoplatform.com">Gennady Azarenkov</a>
 * @version $Id: VersionableWorkspaceDataManager.java 5834 2006-05-29 12:05:59Z peterit $
 */

public class VersionableWorkspaceDataManager extends CacheableWorkspaceDataManager {

  private CacheableWorkspaceDataManager versionDataManager;
  
  public VersionableWorkspaceDataManager(WorkspaceDataContainer dataContainer,
      WorkspaceStorageCacheImpl cache) {
    super(dataContainer, cache);
  }
  
  // called by WorkspaceContainer after repository initialization
  public void setSystemDataManager(DataManager systemDataManager) {

    this.versionDataManager = (CacheableWorkspaceDataManager) systemDataManager;
  }
  
  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.core.WorkspaceDataManager#getChildNodes(org.exoplatform.services.jcr.datamodel.NodeData, boolean)
   */
  @Override
  public List<NodeData> getChildNodesData(NodeData nodeData, boolean forcePersistentRead) throws RepositoryException {
    InternalQPath path = nodeData.getQPath();
    if(isSystemDescendant(path) && !this.equals(versionDataManager)) {
      return versionDataManager.getChildNodesData(nodeData, forcePersistentRead);
    }
    return super.getChildNodesData(nodeData, forcePersistentRead);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.core.WorkspaceDataManager#getChildNodes(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  @Override
  public List<NodeData> getChildNodesData(NodeData nodeData) throws RepositoryException {
    InternalQPath path = nodeData.getQPath();
    if(isSystemDescendant(path) && !this.equals(versionDataManager)) {
      return versionDataManager.getChildNodesData(nodeData, false);
    }
    return super.getChildNodesData(nodeData, false);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.core.WorkspaceDataManager#getChildNodesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  @Override
  public int getChildNodesCount(NodeData nodeData) throws RepositoryException {
    InternalQPath path = nodeData.getQPath();
    if(isSystemDescendant(path) && !this.equals(versionDataManager)) {
      return versionDataManager.getChildNodesCount(nodeData);
    }
    return super.getChildNodesCount(nodeData);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.core.WorkspaceDataManager#getChildProperties(org.exoplatform.services.jcr.datamodel.NodeData, boolean)
   */
  @Override
  public List<PropertyData> getChildPropertiesData(NodeData nodeData, boolean forcePersistentRead) throws RepositoryException {
    InternalQPath path = nodeData.getQPath();
    if(isSystemDescendant(path) && !this.equals(versionDataManager)) {
      return versionDataManager.getChildPropertiesData(nodeData, forcePersistentRead);
    }
    return super.getChildPropertiesData(nodeData, forcePersistentRead);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.core.WorkspaceDataManager#getChildProperties(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  @Override
  public List<PropertyData> getChildPropertiesData(NodeData nodeData) throws RepositoryException {
    InternalQPath path = nodeData.getQPath();
    if(isSystemDescendant(path) && !this.equals(versionDataManager)) {
      return versionDataManager.getChildPropertiesData(nodeData, false);
    }
    return super.getChildPropertiesData(nodeData, false);
  }

  /* (non-Javadoc)
   * @see org.exoplatform.services.jcr.impl.core.WorkspaceDataManager#getChildPropertiesCount(org.exoplatform.services.jcr.datamodel.NodeData)
   */
  @Override
  public int getChildPropertiesCount(NodeData nodeData) throws RepositoryException {
    InternalQPath path = nodeData.getQPath();
    if(isSystemDescendant(path) && !this.equals(versionDataManager)) {
      return versionDataManager.getChildPropertiesCount(nodeData);
    }   
    return super.getChildPropertiesCount(nodeData);
  }

  /**
   * @see org.exoplatform.services.jcr.impl.dataflow.persistent.WorkspaceDataManager#getItemDataByUUID(java.lang.String)
   */
  public ItemData getItemData(InternalQPath qpath) throws RepositoryException {
    if(isSystemDescendant(qpath) && !this.equals(versionDataManager)) {
      return versionDataManager.getItemData(qpath);
    }
    return super.getItemData(qpath);

  }
  
  /**
   * @see org.exoplatform.services.jcr.impl.dataflow.persistent.WorkspaceDataManager#getItemDataByUUID(java.lang.String)
   */
  public ItemData getItemData(String uuid) throws RepositoryException {
    ItemData data = super.getItemData(uuid);
    if(data != null)
      return data;
    else if(!this.equals(versionDataManager)) { 
      // try from version storage if not the same
      data = versionDataManager.getItemData(uuid);
      if(data != null && isSystemDescendant(data.getQPath()))
        return data;
    } 
    return null;
  }

  public synchronized void save(ItemDataChangesLog changesLog) throws RepositoryException,
      InvalidItemStateException {
    List<ItemState> changes = changesLog.getChanges();
    List<ItemState> versionChanges = new ArrayList<ItemState>();
    List<ItemState> nonVersionChanges = new ArrayList<ItemState>();

    log.debug("save ver: " + changesLog.dump());

    for (int i = 0; i < changes.size(); i++) {
      if (isSystemDescendant(changes.get(i).getData().getQPath()) && !this.equals(versionDataManager))
        versionChanges.add(changes.get(i));
      else
        nonVersionChanges.add(changes.get(i));
    }

    if (!versionChanges.isEmpty())
      versionDataManager.save(new ItemDataChangesLog(versionChanges, changesLog.getSession()));

    if (!nonVersionChanges.isEmpty())
      super.save(new ItemDataChangesLog(nonVersionChanges, changesLog.getSession()));

    // auto add for versionable
    // TODO do not use ItemImpl here!
    // TODO don't use session here!
    // if haven't session - then it's initialization start time (we think that) 
    if (changesLog.getSession() != null) {
      List versionStuffs = new ArrayList();
      for (int i = 0; i < changes.size(); i++) {
        if (changes.get(i).getState() == ItemChangeState.ADDED
            || changes.get(i).getState() == ItemChangeState.AUTO_ADDED) {
          ItemImpl item = changesLog.getSession().getTransientNodesManager().getAccessibleItemByUUID(
              changes.get(i).getData().getUUID());
          traverseVersionable(item, versionStuffs);
        }
      }
      
      for (int i = 0; i < versionStuffs.size(); i++) {
        saveVersionHistory((NodeImpl) versionStuffs.get(i));
      }
    }
  }

  private void saveVersionHistory(NodeImpl versionableNode) throws RepositoryException {

    if (versionDataManager.getItemData(versionableNode.getProperty("jcr:versionHistory")
        .getString()) != null)
      return;

    // System.out.println(">>> Versionable 2 "+versionableNode.getPath());

    VersionHistoryImpl versionHistory = createVersionHistory(versionableNode);

    SessionImpl sess = (SessionImpl) versionableNode.getSession();

    ValueFactory valueFactory = sess.getValueFactory();

    versionHistory.createChildProperty("jcr:primaryType", new Value[] { valueFactory.createValue(
        "nt:versionHistory", PropertyType.NAME) }, PropertyType.NAME);
    versionHistory.createChildProperty("jcr:uuid", new Value[] { valueFactory
        .createValue(versionableNode.getProperty("jcr:versionHistory").getString()) },
        PropertyType.STRING);
    versionHistory.createChildProperty("jcr:versionableUuid", new Value[] { valueFactory
        .createValue(versionableNode) }, PropertyType.REFERENCE);

    // Version labels
    NodeImpl versionLabels = versionHistory.createChildNode("jcr:versionLabels",
        "nt:versionLabels", true);

    VersionImpl rootVersion = versionHistory.addRootVersion(versionableNode);

    TransientNodeData data = (TransientNodeData) versionHistory.getActualItemData();
    List<ItemState> changes = new ArrayList<ItemState>();
    traverseVersionHistory(data, changes);

    versionDataManager.save(new ItemDataChangesLog(changes, (SessionImpl) versionHistory
        .getSession()));
  }

  
  private void traverseVersionHistory(TransientItemData data, List changes)
      throws RepositoryException {
    changes.add(new ItemState(data, ItemChangeState.ADDED));
    if (data.isNode()) {
      TransientNodeData nodeData = (TransientNodeData) data;
      List props = nodeData.getChildProperties();               
      for (int i = 0; i < props.size(); i++) {
        traverseVersionHistory((TransientItemData) props.get(i), changes);
      }
      List nodes = nodeData.getChildNodes();
      for (int i = 0; i < nodes.size(); i++) {
        traverseVersionHistory((TransientItemData) nodes.get(i), changes);
      }
    }
  }  
  
  private void traverseVersionable(ItemImpl item, List changes) throws RepositoryException {
    if (item.isNode() && ((NodeImpl)item).isNodeType("mix:versionable")) {
      changes.add(item);
    } else if (!item.isNode() && item.getName().equals("jcr:versionHistory")) {
      // mix:versionable seems to be assigned
      String parentUuid = item.getActualItemData().getParentUUID(); 
      //addVersionHistory((NodeImpl)item.getSession().getNodeByUUID(parentUuid));
      changes.add((NodeImpl)item.getSession().getNodeByUUID(parentUuid));
    }
    
    if(item.isNode()) {
      for (NodeIterator nodes = ((NodeImpl)item).getNodes(); nodes.hasNext();) {
        traverseVersionable((NodeImpl) nodes.nextNode(), changes);
      }
    }
  }
  
  /**
   * @deprecated
   * @param path
   * @return
   */
  private boolean isSystemDescendant(JCRPath path) {
    return path.getAsString(false).equals("/jcr:system") || path.getInternalPath().isDescendantOf(Constants.JCR_SYSTEM_PATH, false);

  }
  
  private boolean isSystemDescendant(InternalQPath path) {
    return path.equals(Constants.JCR_SYSTEM_PATH) || path.isDescendantOf(Constants.JCR_SYSTEM_PATH, false);
  }

  
  /**
   * Initializes version history for mix:versionable node
   * 
   * @param versionableNode
   * @throws RepositoryException
   */
  private void addVersionHistory(NodeImpl versionableNode) throws RepositoryException {
    
    
    if(versionDataManager.getItemData(versionableNode.
        getProperty("jcr:versionHistory").getString()) != null)
        return;
    
    //System.out.println(">>> Versionable 2 "+versionableNode.getPath());
    
    VersionHistoryImpl versionHistory = createVersionHistory(versionableNode);

    SessionImpl sess = (SessionImpl)versionableNode.getSession();
    
    ValueFactory valueFactory = sess.getValueFactory();

    versionHistory.createChildProperty("jcr:primaryType",
        new Value[] { valueFactory.createValue("nt:versionHistory",
            PropertyType.NAME) }, PropertyType.NAME);
    versionHistory.createChildProperty("jcr:uuid", new Value[] { valueFactory
        .createValue(versionableNode.getProperty("jcr:versionHistory").getString()) },
        PropertyType.STRING);
    versionHistory.createChildProperty("jcr:versionableUuid",
        new Value[] { valueFactory.createValue(versionableNode) }, PropertyType.REFERENCE);

    // Version labels
    NodeImpl versionLabels = versionHistory.createChildNode(
        "jcr:versionLabels", "nt:versionLabels", true);
    
    versionHistory.addRootVersion(versionableNode);
    
    TransientNodeData data = (TransientNodeData) versionHistory.getActualItemData();
    List<ItemState> changes = new ArrayList<ItemState>();
    traverseVersionHistory(data, changes);
    versionDataManager.save(new ItemDataChangesLog(changes, (SessionImpl) versionHistory
        .getSession()));

    // Save
    // [PN] 07.04.06
    //versionDataManager.add(versionHistory);
    //((SessionImpl)versionHistory.getSession()).getTransientNodesManager().getTransactManager().saveItem(versionHistory);

  }
  
  private VersionHistoryImpl createVersionHistory(Node versionableNode)
      throws RepositoryException {
    
    SessionImpl sess = (SessionImpl)versionableNode.getSession();
    NodeImpl versionStorage = versionStorageNode(sess);

    // Create new NodeData inmemory
    JCRPath path = sess.getLocationFactory().createJCRPath(versionStorage.getLocation(),
        versionableNode.getUUID());
    String uuid = versionableNode.getProperty("jcr:versionHistory").getString();
    int orderNum = 0;
    NodeData nodeData = new TransientNodeData(path.getInternalPath(), uuid, -1,
        Constants.NT_VERSION_HISTORY, new InternalQName[0], orderNum,
        versionStorage.getInternalUUID());
    
    // [PN] 07.04.06 used sess instead versionableNode.getSession()
    VersionHistoryImpl versionHistory = new VersionHistoryImpl(nodeData,
        versionStorage, versionStorage.getState(), sess);
    
    return versionHistory;
  }
  
  
  private NodeImpl versionStorageNode(SessionImpl session) throws RepositoryException {
    
    JCRPath path = session.getLocationFactory().parseAbsPath("/jcr:system/jcr:versionStorage");
    NodeData data = (NodeData) getItemData(path.getInternalPath());
    return new NodeImpl(data, null, ItemChangeState.UNCHANGED, session);
  }  
}
