/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/
package org.exoplatform.services.jcr.impl.dataflow.session;

import java.util.ArrayList;
import java.util.List;

import javax.jcr.ItemExistsException;
import javax.jcr.RepositoryException;

import org.exoplatform.services.jcr.datamodel.InternalQPath;
import org.exoplatform.services.jcr.datamodel.ItemData;
import org.exoplatform.services.jcr.datamodel.NodeData;
import org.exoplatform.services.jcr.impl.core.ItemChangeState;
import org.exoplatform.services.jcr.impl.core.ItemImpl;
import org.exoplatform.services.jcr.impl.core.ItemState;
import org.exoplatform.services.jcr.impl.core.NodeImpl;
import org.exoplatform.services.jcr.impl.core.PropertyImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.dataflow.ItemDataChangesLog;
import org.exoplatform.services.jcr.impl.dataflow.TransientItemData;
import org.exoplatform.services.jcr.impl.dataflow.persistent.BatcheableWorkspaceDataManager;


/**
 * Created by The eXo Platform SARL        .
 *
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov</a>
 * @version $Id: SessionDataflowManager.java 5800 2006-05-28 18:03:31Z geaz $
 * 
 * Responsible for session items data flow. methods never should be called
 * by api implementation classes (like ItemImpl etc) directly!
 */

public class SessionDataflowManager {
  
  private final SessionChangesLog changesLog;
  private final SessionItemPool itemsPool;
  private final TransactionableDataManager transactManager;
  private final ItemFactory itemFactory;
  private final SessionImpl session;

  public SessionDataflowManager(BatcheableWorkspaceDataManager storageDataManager, 
      SessionImpl session) {
    changesLog = new SessionChangesLog(session);
    transactManager = new TransactionableDataManager(storageDataManager,
        session);
    itemsPool = new SessionItemPool();
    itemFactory = new ItemFactory(session);
    this.session = session;
  }

  
  public ItemDataChangesLog getChanges(InternalQPath rootPath) {
    List <ItemState> changes = changesLog.getChanges(rootPath);
    return new ItemDataChangesLog(changes, session);
  }
  
  public ItemImpl add(TransientItemData data, boolean forcePooling) throws RepositoryException,
    ItemExistsException {
    
    ItemImpl item = itemsPool.get(data.getQPath());
    if(item == null) {
      item = itemFactory.createItem(data);
      if(forcePooling)
        itemsPool.put(item);
    } else {
      if(item.getState() == ItemChangeState.DELETED)
        item.resetUUID(data.getUUID());
      else
        throw new ItemExistsException("Item "+item.getPath()+" already exists");
    }
    
    ItemState change = new ItemState(data, ItemChangeState.ADDED);
    changesLog.add(change);
    
    return item;
  }
  
  public ItemImpl update(TransientItemData data, boolean forcePooling) throws RepositoryException {
    
    ItemImpl item = itemsPool.get(data.getQPath());
    if(item == null) {
      item = itemFactory.createItem(data);
      if(forcePooling)
        itemsPool.put(item);
    } else {
      if(item.getState() == ItemChangeState.DELETED)
        throw new ItemExistsException("Item "+item.getPath()+" is deleted");
    }
    
    ItemState change = new ItemState(data, ItemChangeState.UPDATED);
    changesLog.add(change);
    return item;
    
  }
  
  public void delete(TransientItemData data) throws RepositoryException {
    
    ItemState change = new ItemState(data, ItemChangeState.DELETED);
    
    changesLog.add(change);
    ItemImpl item = itemFactory.createItem(data);
    itemsPool.remove(data.getUUID());
    
  }
  
  /**
   * Merges data 
   * @param data
   * @return
   * @throws RepositoryException
   * @throws IllegalStateException
   */
  public ItemImpl refresh(ItemData data) throws RepositoryException,
    IllegalStateException {
  
  ItemImpl item = itemsPool.get(data.getQPath());
  if(item == null) {
    item = itemFactory.createItem(data);
    //itemsPool.put(item);
  } else {
    if(item.getState() == ItemChangeState.DELETED)
      throw new IllegalStateException("Item "+item.getPath()+" is deleted");
    item.resetUUID(data.getUUID());
    
  }
  
  ItemState change = new ItemState(data, ItemChangeState.ADDED);
  changesLog.add(change);
  
  return item;
}

  
  public void clean(InternalQPath rootPath) {
    changesLog.remove(rootPath);
  }

  public final void save(InternalQPath rootPath) throws RepositoryException {
    transactManager.save(getChanges(rootPath));
    changesLog.remove(rootPath);
  }
  
  public List <NodeImpl> getChildNodes(NodeData parent) throws RepositoryException {
    List <NodeData> childData =
      transactManager.getChildNodesData(parent);
    List <NodeImpl> storedChildren = new ArrayList <NodeImpl> ();
    for(int i=0; i<childData.size(); i++) {
      NodeData data = childData.get(i);
      storedChildren.add(itemFactory.createNode(data));
    }
    List <NodeImpl> nodes = mergeNodes(itemsPool.getChildNodes(parent.getUUID()), 
        storedChildren);
    for(int i=0; i<nodes.size(); i++) {
      itemsPool.put(nodes.get(i));
    }
    return nodes;
  }

  public List <PropertyImpl> getChildProperties(NodeData parent) throws RepositoryException {
    List <NodeData> childData =
      transactManager.getChildNodesData(parent);
    List <PropertyImpl> storedChildren = new ArrayList <PropertyImpl> ();
    for(int i=0; i<childData.size(); i++) {
      NodeData data = childData.get(i);
      storedChildren.add(itemFactory.createProperty(data));
    }
    List <PropertyImpl> props = mergeProperties(itemsPool.getChildProperties(parent.getUUID()), 
        storedChildren);
    for(int i=0; i<props.size(); i++) {
      itemsPool.put(props.get(i));
    }
    return props;
  }
  
  public ItemImpl getItem(InternalQPath path) throws RepositoryException {
    ItemImpl item = itemsPool.get(path);
    if(item == null) {
      ItemData data = transactManager.getItemData(path);
      if(data != null) { 
        item = itemFactory.createItem(data); 
        itemsPool.put(item);
        return item;
      } else
        return null;
    } else {
      return item;
    }
  }

  public ItemImpl getItem(String uuid) throws RepositoryException {
    ItemImpl item = itemsPool.get(uuid);
    if(item == null) {
      ItemData data = transactManager.getItemData(uuid); 
      if(data != null) {
        item = itemFactory.createItem(data); 
        itemsPool.put(item);
        return item;
      } else
        return null;
    } else {
      return item;
    }
  }
  

  /**
   * Merges 2 lists of itemImpl, if both contain the same item - recent added
   * @param recentData
   * @param olderData
   * @return
   */
  protected final List <NodeImpl> mergeNodes(List <NodeImpl> recentData, List <NodeImpl> olderData) {

    List <NodeImpl> list = new ArrayList<NodeImpl>();
    if(olderData.size() == 0) {
      list.addAll(recentData);
    } else if(recentData.size() == 0) {
      list.addAll(olderData);
    } else {
      list.addAll(recentData);
      for (int i = 0; i < olderData.size(); i++) {
        boolean contain = false;
        for (int j = 0; j < recentData.size(); j++) {
          if (olderData.get(i).getInternalUUID().equals(recentData.get(j).getInternalUUID())) {
            contain = true;
          }
        }
        if (!contain)
          list.add(olderData.get(i));
      }
    }
    return list;
  }

  
  protected final List <PropertyImpl> mergeProperties(List <PropertyImpl> recentData, List <PropertyImpl> olderData) {

    List <PropertyImpl> list = new ArrayList<PropertyImpl>();
    if(olderData.size() == 0) {
      list.addAll(recentData);
    } else if(recentData.size() == 0) {
      list.addAll(olderData);
    } else {
      list.addAll(recentData);
      for (int i = 0; i < olderData.size(); i++) {
        boolean contain = false;
        for (int j = 0; j < recentData.size(); j++) {
          if (olderData.get(i).getInternalUUID().equals(recentData.get(j).getInternalUUID())) {
            contain = true;
          }
        }
        if (!contain)
          list.add(olderData.get(i));
      }
    }
    return list;
  }

}
