/***************************************************************************
 * Copyright 2001-2003 The eXo Platform SARL         All rights reserved.  *
 * Please look at license.txt in info directory for more license detail.   *
 **************************************************************************/

package org.exoplatform.services.jcr.impl.xml;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import javax.jcr.ImportUUIDBehavior;
import javax.jcr.ItemExistsException;
import javax.jcr.NamespaceRegistry;
import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.PathNotFoundException;
import javax.jcr.PropertyType;
import javax.jcr.RepositoryException;
import javax.jcr.Value;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.PropertyDefinition;

import org.exoplatform.services.jcr.datamodel.InternalQName;
import org.exoplatform.services.jcr.impl.Constants;
import org.exoplatform.services.jcr.impl.core.NodeImpl;
import org.exoplatform.services.jcr.impl.core.PropertyImpl;
import org.exoplatform.services.jcr.impl.core.SessionImpl;
import org.exoplatform.services.jcr.impl.util.ISO9075;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;

/**
 * Created by The eXo Platform SARL .
 * 
 * @author <a href="mailto:geaz@users.sourceforge.net">Gennady Azarenkov</a>
 * @version $Id: DocNodeImporter.java 5307 2006-04-27 13:55:57Z peterit $
 */

class DocNodeImporter implements ContentHandler {

  private SessionImpl       session;

  private NodeImpl          rootNode;
  
  private NamespaceRegistry nsReg;

  private Stack             tree;
  
  private int uuidBehavior = -1;
  
  private Exception causeException = null;
  
  private boolean saveOnEnd = false;

  DocNodeImporter(NodeImpl parent, int uuidBehavior) throws RepositoryException {
    //super(parent);
    this.tree = new Stack();
    this.session = (SessionImpl) parent.getSession();
    this.uuidBehavior = uuidBehavior;
    this.nsReg = session.getWorkspace().getNamespaceRegistry();
    tree.push(parent);
  }

  /*
   * (non-Javadoc)
   * 
   * @see org.exoplatform.services.jcr.impl.xml.NodeImporter#startElement(java.lang.String,
   *      java.lang.String, java.lang.String, org.xml.sax.Attributes)
   */
  public void startElement(String namespaceURI, String localName, String qName, Attributes atts)
      throws SAXException {

    //System.err.println("================ startElement:" + localName);
    
    if ("jcr:root".equals(qName))
      qName = "";

    HashMap<String, String> props = new HashMap<String, String>();
    // default value of primary type
    String primaryNodeType = "nt:unstructured";
    props.put("jcr:primaryType", primaryNodeType);

    String[] mixinNodeTypes = null;

    if (atts != null) {
      int length = atts.getLength();
      for (int i = 0; i < length; i++) {
        String propName;
        String attrQName = atts.getQName(i);
        if (!"".equals(attrQName))
          propName = attrQName;
        else
          propName = atts.getLocalName(i);

        if (propName.equals("jcr:primaryType")) {
          primaryNodeType = atts.getValue(i);

          props.put(propName, atts.getValue(i));
        } else if (propName.equals("jcr:mixinTypes")) {
          mixinNodeTypes = atts.getValue(i).split(" ");
          for (int mi = 0; mi < mixinNodeTypes.length; mi++) {
            mixinNodeTypes[mi] = ISO9075.decode(mixinNodeTypes[mi]);
          }

          props.put(propName, atts.getValue(i));
        } else if (propName.equals("jcr:uuid")) {
          props.put(propName, atts.getValue(i));
        } else {
          props.put(propName, atts.getValue(i));
        }
      }
    }

    if (mixinNodeTypes == null) {
      mixinNodeTypes = new String[0];
    }

    // fill mixinNodeTypes array
    boolean hasMixReferenceable = false;
    for (int i = 0; i < mixinNodeTypes.length; i++) {
      if ("mix:referenceable".equals(mixinNodeTypes[i])) {
        hasMixReferenceable = true;
      }
    }

    NodeImpl node = null;
    NodeImpl sameUuidNode = null;
    String uuid = props.get("jcr:uuid");
    try {

      NodeImpl parentNode = parent();

      if (hasMixReferenceable && uuid != null) {

        // Check for mix:referenceable node with uuid attr
        sameUuidNode = (NodeImpl) session.getNodeByUUID(uuid);
        if (sameUuidNode != null) {
          switch (uuidBehavior) {
          case ImportUUIDBehavior.IMPORT_UUID_CREATE_NEW:
            // Incoming referenceable nodes are assigned newly created UUIDs
            // upon addition to the workspace. As a result UUID collisions
            // never occur.

            // reset UUID and it will be autocreated in session
            //System.err.println("====== [IMPORT_UUID_CREATE_NEW] " + sameUuidNode.getPath() + " " + sameUuidNode.getInternalUUID());
            uuid = null;
            break;
          case ImportUUIDBehavior.IMPORT_UUID_COLLISION_REMOVE_EXISTING:
            // If an incoming referenceable node has the same UUID as a node
            // already existing in the workspace then the already existing
            // node (and its subtree) is removed from wherever it may be in
            // the workspace before the incoming node is added. Note that this
            // can result in nodes �disappearing� from locations in the
            // workspace that are remote from the location to which the
            // incoming subtree is being written.

            // IMPORT_UUID_COLLISION_REMOVE_EXISTING");
            NodeIterator samePatterns = sameUuidNode.getNodes(parentNode.getName());
            if (samePatterns.hasNext()) {
              // Throw exception
              throw new ConstraintViolationException(
                  "A uuidBehavior is set to IMPORT_UUID_COLLISION_REMOVE_EXISTING and an incoming node has the same UUID as the node at parentAbsPath or one of its ancestors");
            } else {
              // remove node
              // Node sameUuidParent = sameUuidNode.getParent();
              sameUuidNode.remove();
              sameUuidNode.save();
              //System.err.println("====== remove sameUuidNode [IMPORT_UUID_COLLISION_REMOVE_EXISTING]: " + sameUuidNode.getPath() + " " + sameUuidNode.getInternalUUID());
              sameUuidNode = null;
              // sameUuidParent.save();
            }
            break;
          case ImportUUIDBehavior.IMPORT_UUID_COLLISION_REPLACE_EXISTING:
            // If an incoming referenceable node has the same UUID as a node
            // already existing in the workspace, then the already existing
            // node is replaced by the incoming node in the same position as
            // the existing node. Note that this may result in the incoming
            // subtree being disaggregated and �spread around� to different
            // locations in the workspace. In the most extreme case this
            // behavior may result in no node at all being added as child of
            // parentAbsPath. This will occur if the topmost element of the
            // incoming XML has the same UUID as an existing node elsewhere in
            // the workspace.

            // replace in same location
            parentNode = (NodeImpl) sameUuidNode.getParent();
            sameUuidNode.remove();
            sameUuidNode.save();
            //System.err.println("====== remove sameUuidNode [IMPORT_UUID_COLLISION_REPLACE_EXISTING]: " + sameUuidNode.getPath() + " " + sameUuidNode.getInternalUUID());
            sameUuidNode = null;
            break;
          case ImportUUIDBehavior.IMPORT_UUID_COLLISION_THROW:
            // If an incoming referenceable node has the same UUID as a node
            // already existing in the workspace then a SAXException is thrown
            // by the ContentHandler during deserialization.
            throw new ItemExistsException(
                "An incoming referenceable node has the same UUID as a node already existing in the workspace!");
          default:
          }
        }
      } else {
        // other cases with same names etc.

        try {
          // Check for node with this name in session
          NodeImpl sameNameNode = (NodeImpl) parentNode.getNode(qName);
          if (!sameNameNode.getDefinition().allowsSameNameSiblings()) {
            // Throw exception
            String msg = "A node with this name (" + qName + ") is already exists. "
                + "Overwriting an existing node during import is impossible.";
            throw new SAXException(msg);
          }
        } catch (PathNotFoundException e) {
          // ok
        }
      }

      // Code before 20.12.2005, Peter Nedonosko
      // node = parent().createChildNode(qName, primaryNodeType, mixinNodeTypes,
      // false); // addAutoCreated=false

      node = parentNode.createChildNode(qName, primaryNodeType, mixinNodeTypes, uuid, true, false);
      // autoCreated=true,
      // addAutoCreated=false
      
      //System.err.println("====== uuid: " + node.getInternalUUID());

      Iterator<String> keys = props.keySet().iterator();
      while (keys.hasNext()) {
        String key = keys.next();

        if (key.equals("jcr:primaryType")) {
          Value value = session.getValueFactory().createValue(props.get(key),
              PropertyType.NAME);
          node.createChildPropertyInmemory(key, PropertyType.NAME, new Value[] { value }, false, false);
        } else if (key.equals("jcr:mixinTypes")) {
          Value[] values = new Value[mixinNodeTypes.length];
          for (int vi = 0; vi < mixinNodeTypes.length; vi++) {
            Value value = session.getValueFactory().createValue(props.get(key),
                PropertyType.NAME);
            values[vi] = value;
          }
          node.createChildPropertyInmemory(key, PropertyType.NAME, values, false, false);
        } else if (hasMixReferenceable && key.equals("jcr:uuid")) {
          // referenceable node and uuid value
          // [PN] 31.03.06 It's wrong, jcr:uuid has PropertyType.STRING see 6.7.21.2 mix:referenceable
          //Value value = session.getValueFactory().createValue((String) props.get(key), PropertyType.REFERENCE);
          //node.createChildPropertyInmemory(key, PropertyType.REFERENCE, new Value[] { value }, false, false);
          Value value = session.getValueFactory().createValue(node.getInternalUUID(), PropertyType.STRING);
          //Value value = session.getValueFactory().createValue(props.get(key), PropertyType.STRING);
          node.createChildPropertyInmemory(key, PropertyType.STRING, new Value[] { value }, false, false);
          //System.err.println("====== jcr:uuid: " + props.get(key) + " " + node.getInternalUUID() + " " + node.getProperty("jcr:uuid").getString());
        } else {
          // string value!
          Value value = session.getValueFactory().createValue(props.get(key));
          // [PN] 09.03.06 Multy-valued value as DEFAULT will be used if no definition found on node
          //PropertyDefinition pd = node.findPropertyDefinition(key);
          //createChildProperty(node, key, PropertyType.STRING, new Value[] { value }, 
          //    (pd != null ? pd.isMultiple() : true)); 
          node.createChildPropertyInmemory(key, PropertyType.STRING, new Value[] { value }, false, false);
        }
      }

    } catch (Exception e) {
      e.printStackTrace();
      throw new SAXException(e.getMessage(), e);
    }
    if (rootNode == null) {
      rootNode = node;
    }

    tree.push(node);
  }

  public void endElement(String namespaceURI, String localName, String qName) throws SAXException {
    NodeImpl node = (NodeImpl) tree.pop();
    if (!isSaveOnEnd()) {
      // [PN] 03.02.06 Put this node in changedStorage in SDM for save method in Session/Node
      try {
        session.getTransientNodesManager().update(node); // update in changedStorage
        session.getTransientNodesManager().removeFromCache(node);
        // [PN] 15.04.06 Cache is valid always
        //session.getTransientNodesManager().getWorkspaceDataManager().removeFromCache(node.getActualItemData());
      } catch (Exception e) {
        System.err.println("[ERROR] " + 
            getClass().getName() + ".endElement() node import error: " + e.getMessage());
      }
    }
  }

  public void characters(char[] ch, int start, int length) throws SAXException {
    // System.out.println("chars "+new String(ch)+" "+length+" "+start);
    
    // [PN] 27.04.06 StringBuilder instead StringBuffer
    StringBuilder text = new StringBuilder();  
    text.append(ch, start, length);

    try {
      NodeImpl txtNode = parent().createChildNode("jcr:xmltext", "nt:unstructured", new String[0],
          false);
      txtNode.createChildPropertyInmemory(Constants.JCR_PRIMARYTYPE, PropertyType.NAME, 
          new Value[] { session.getValueFactory().createValue("nt:unstructured", PropertyType.NAME) },
          false, false);
      txtNode.createChildPropertyInmemory(Constants.JCR_XMLCHARACTERS, PropertyType.STRING,
          // [PN] 27.04.06 text.toString() instead new String(text)
          new Value[] { session.getValueFactory().createValue(text.toString(), PropertyType.STRING) },
          false, false, false); //6.4.2.3 a single-valued property called jcr:xmlcharacters,
    } catch (RepositoryException e) {
      throw new RuntimeException(e);
    }
  }

  public void setDocumentLocator(Locator locator) {
  }

  public void startDocument() throws SAXException {
  }

  public void endDocument() throws SAXException {
  }

  public void startPrefixMapping(String prefix, String uri) throws SAXException {
  }

  public void endPrefixMapping(String prefix) throws SAXException {
  }

  public void ignorableWhitespace(char[] ch, int start, int length) throws SAXException {
  }

  public void processingInstruction(String target, String data) throws SAXException {
  }

  public void skippedEntity(String name) throws SAXException {
  }

  private NodeImpl parent() {
    return (NodeImpl) tree.peek();
  }
  
  public boolean isSaveOnEnd() {
    return saveOnEnd;
  }

  public void setSaveOnEnd(boolean saveOnEnd) {
    this.saveOnEnd = saveOnEnd;
  }
  
  NodeImpl getNode() {
    return rootNode;
  }

}
